
#include "connectivity.h"
#include "Map.h"
#include <cstring>
#include <cstdlib>
#include <sstream>
#include <iomanip>
#include <string>

namespace cnct
{

std::vector<Node*> Node::cache;
#ifdef MEMORY_CHECK
int Node::allocated = 0;

int Node::unreturned()
{
    return allocated - cache.size();
}
#endif

Node* Node::allocate()
{
    Node* n;
    if (cache.size() != 0)
    {
        n = cache.back();
        cache.pop_back();
#ifdef MEMORY_CHECK
        n->area = Territory::allocate();
#endif
    } else {
        n = new Node();
#ifdef MEMORY_CHECK
        ++allocated;
#endif
    }
    n->clear();
    return n;
}

Node* Node::free(Node* node)
{
    int num_children = node->children.size();
    for (int i=0; i < num_children; i++)
    {
        Node::free(node->children[i].node);
    }
    node->children.clear();
#ifdef MEMORY_CHECK
    node->area = Territory::free(node->area);
#endif
    cache.push_back(node);
    return NULL;
}

Node::Node() : parent(NULL), age(0), my_size(0)
{
    entrance[0] = 0;
    entrance[1] = 0;
    area = Territory::allocate();
}

Node::~Node()
{
    area = Territory::free(area);
    int num_children = children.size();
    for (int i=0; i < num_children; i++)
    {
        Node::free(children[i].node);
    }
    children.clear();
}

void Node::clear()
{
    parent = NULL;
    entrance[0] = 0;
    entrance[1] = 0;
    doubled_entrance = false;
    area->clear();
    my_size = 0;
    age = 0;
    if (children.size() != 0)
    {
        int num_children = children.size();
        for (int i=0; i < num_children; i++)
        {
            Node::free(children[i].node);
        }
        children.clear();
    }
}

int Node::set_size()
{
    my_size = area->size(entrance);
    my_bound = my_size == 3 ? EXACT : UPPER;
    if (children.size() != 0)
    {
        std::vector<ChildInfo>::iterator iter;
        for (iter=children.begin(); iter < children.end(); iter++)
        {
            Node* node = iter->node;
            node->set_size();
            int to_child;
            if (!doubled_entrance && (entrance[0] == iter->exit[0]
                        && entrance[1] == iter->exit[1]))
            {
                to_child = 1;
            } else {
                to_child = area->size(entrance, iter->exit);
            }
            iter->best_size = node->my_size + to_child;
            iter->bound = node->my_bound == UPPER ? UPPER
                : to_child == 3 ? EXACT : UPPER;
            if (iter->best_size > my_size)
            {
                my_size = iter->best_size;
                my_bound = iter->bound;
            }
        }
    }
    return my_size;
}

void Tree::collapse_doubled(Node* node)
{
    if (node->children.size() == 1 && node->children[0].node->doubled_entrance
            && node->area->size(node->entrance) == 1)
    {
        Node* second = node->children[0].node;
        *node->area |= *second->area;
        node->children.clear();
        for (int i=0; i < second->children.size(); i++)
        {
            node->children.push_back(second->children[i]);
            second->children[i].node->parent = node;
        }
        second->children.clear();
        TerritoryIter sp_iter = TerritoryIter(second->area);
        while (sp_iter.has_more())
        {
            int* loc = sp_iter.next();
            node_map[loc[0]][loc[1]] = node;
        }
        second = Node::free(second);
    }
    for (int i=0; i < node->children.size(); i++)
    {
		collapse_doubled(node->children[i].node);
    }
}

std::vector<Node***> Tree::map_cache;
#ifdef MEMORY_CHECK
int Tree::allocated = 0;

int Tree::unreturned()
{
    return allocated - map_cache.size();
}
#endif

Node*** Tree::allocate_map()
{
    Node*** n;
    if (map_cache.size() != 0)
    {
        n = map_cache.back();
        map_cache.pop_back();
    } else {
        n = new Node**[Map::map_width];
        for (int i=0; i < Map::map_width; i++)
        {
            n[i] = new Node*[Map::map_height];
        }
#ifdef MEMORY_CHECK
        ++allocated;
#endif
    }
    for (int i=0; i < Map::map_width; i++)
    {
        memset(n[i], NULL, sizeof(Node*) * Map::map_height);
    }
    return n;
}

Node*** Tree::free_map(Node*** map)
{
    map_cache.push_back(map);
    return NULL;
}

Node* Tree::combine_nodes(Node* first, Node* second)
{
    ++age;
    Node* branch = first;
    while (branch != second && branch != NULL)
    {
        branch->age = age;
        branch = branch->parent;
    }
    if (branch != second)
    {
        branch = second;
        while (branch->age != age && branch != NULL)
        {
            branch->age = age;
            branch = branch->parent;
        }
        if (branch == NULL)
        {
            fprintf(stderr, "Tried to combine nodes from separate trees.\n");
            exit(1);
        }
    }
    if (branch != first)
    {
        while (branch != first->parent)
        {
            Node* parent = first->parent;
            *parent->area |= *first->area;
            for (int i=0; i < parent->children.size(); i++)
            {
                if (parent->children[i].node == first)
                {
                    parent->children.erase(parent->children.begin()+i);
                    break;
                }
            }
            for (int i=0; i < first->children.size(); i++)
            {
                parent->children.push_back(first->children[i]);
                first->children[i].node->parent = parent;
            }
            first->children.clear();
            first = Node::free(first);
            first = parent;
        }
    }
    if (branch != second)
    {
        while (branch != second->parent)
        {
            Node* parent = second->parent;
            *parent->area |= *second->area;
            for (int i=0; i < parent->children.size(); i++)
            {
                if (parent->children[i].node == second)
                {
                    parent->children.erase(parent->children.begin()+i);
                    break;
                }
            }
            for (int i=0; i < second->children.size(); i++)
            {
                parent->children.push_back(second->children[i]);
                second->children[i].node->parent = parent;
            }
            second->children.clear();
            second = Node::free(second);
            second = parent;
        }
    }
    Node* out_node = branch;
    bool merged_to_branch = false;
    if (branch != first && branch != second)
    {
        bool same_exit = false;
        bool already_doubled = first->doubled_entrance || second->doubled_entrance;
        int first_ci = -1;
        int second_ci = -1;
        for (int i=0; i < branch->children.size(); i++)
        {
            if (branch->children[i].node == first)
            {
                first_ci = i;
                if (second_ci != -1)
                    break;
            }
            else if (branch->children[i].node == second)
            {
                second_ci = i;
                if (first_ci != -1)
                    break;
            }
        }
        if (branch->children[first_ci].exit[0] == branch->children[second_ci].exit[0]
                && branch->children[first_ci].exit[1] == branch->children[second_ci].exit[1])
            same_exit = true;
        // merge first and second together
        *first->area |= *second->area;
        branch->children.erase(branch->children.begin()+second_ci);
        for (int i=0; i < second->children.size(); i++)
        {
            first->children.push_back(second->children[i]);
            second->children[i].node->parent = first;
        }
        second->children.clear();
        second = Node::free(second);
        first->doubled_entrance = true;
        // merge into parent if not cut node
        if (already_doubled || !same_exit)
        {
            *branch->area |= *first->area;
            for (int i=0; i < branch->children.size(); i++)
            {
                if (branch->children[i].node == first)
                {
                    // can't use first_ci because it is invalidated when erasing second_ci
                    branch->children.erase(branch->children.begin()+i);
                    break;
                }
            }
            for (int i=0; i < first->children.size(); i++)
            {
                branch->children.push_back(first->children[i]);
                first->children[i].node->parent = branch;
            }
            first->children.clear();
            first = Node::free(first);
            merged_to_branch = true;
        } else {
            for (int x=0; x < Map::map_width; x++)
            {
                uint64_t column = (*first->area)[x];
                while (column)
                {
                    uint64_t cbit = column & -column;
                    column ^= cbit;
                    int y = bitindex(cbit);
                    node_map[x][y] = first;
                }
            }
            /*
            TerritoryIter fr_iter = TerritoryIter(first->area);
            while (fr_iter.has_more())
            {
                int* loc = fr_iter.next();
                node_map[loc[0]][loc[1]] = first;
            }
            */
            out_node = first;
        }
    } else {
        if (branch != first)
        {
            second = first;
        }
        *branch->area |= *second->area;
        for (int i=0; i < branch->children.size(); i++)
        {
            if (branch->children[i].node == second)
            {
                branch->children.erase(branch->children.begin()+i);
                break;
            }
        }
        for (int i=0; i < second->children.size(); i++)
        {
            branch->children.push_back(second->children[i]);
            second->children[i].node->parent = branch;
        }
        second->children.clear();
        second = Node::free(second);
        merged_to_branch = true;
    }

    if (merged_to_branch)
    {
        for (int x=0; x < Map::map_width; x++)
        {
            uint64_t column = (*branch->area)[x];
            while (column)
            {
                uint64_t cbit = column & -column;
                column ^= cbit;
                int y = bitindex(cbit);
                node_map[x][y] = branch;
            }
        }
        /*
        TerritoryIter br_iter = TerritoryIter(branch->area);
        while (br_iter.has_more())
        {
            int* loc = br_iter.next();
            node_map[loc[0]][loc[1]] = branch;
        }
        */
    }
    return out_node;
}

Tree::Tree(Territory& terr, int* start) : age(0)
{
    root = Node::allocate();
    root->entrance[0] = start[0];
    root->entrance[1] = start[1];
    area = Territory::allocate();
    (*area)[start[0]] |= UINT64_C(1) << start[1];
    (*root->area)[start[0]] |= UINT64_C(1) << start[1];
    node_map = allocate_map();
    node_map[start[0]][start[1]] = root;
    Territory* explore = Territory::allocate(*area);
    explore->expand();
    explore->andnot(*area);
    *explore &= terr;
    while (explore->not_empty())
    {
        TerritoryIter ex_iter = TerritoryIter(explore);
        while (ex_iter.has_more())
        {
            int* loc = ex_iter.next();
            Node* neighbor = NULL;
            int neighbor_exit[2];
            bool start_node = true;
            for (int dir=0; dir < 4; dir++)
            {
                int x, y;
                switch (dir)
                {
                    case 0:
                        x = loc[0];
                        y = loc[1]-1;
                        break;
                    case 1:
                        x = loc[0]+1;
                        y = loc[1];
                        break;
                    case 2:
                        x = loc[0];
                        y = loc[1]+1;
                        break;
                    case 3:
                        x = loc[0]-1;
                        y = loc[1];
                        break;
                }
                if (!terr.contains(x, y))
                    continue;
                if (node_map[x][y] != NULL)
                {
                    if (neighbor == NULL)
                    {
                        neighbor = node_map[x][y];
                        neighbor_exit[0] = x;
                        neighbor_exit[1] = y;
                    } else {
                        if (node_map[x][y] != neighbor)
                        {
                            neighbor = combine_nodes(neighbor, node_map[x][y]);
                        }
                        (*neighbor->area)[loc[0]] |= UINT64_C(1) << loc[1];
                        (*area)[loc[0]] |= UINT64_C(1) << loc[1];
                        node_map[loc[0]][loc[1]] = neighbor;
                        start_node = false;
                    }
                }
            }
            if (start_node)
            {
                Node* new_node = Node::allocate();
                new_node->parent = neighbor;
                new_node->entrance[0] = loc[0];
                new_node->entrance[1] = loc[1];
                (*new_node->area)[loc[0]] |= UINT64_C(1) << loc[1];
                ChildInfo cinfo = ChildInfo(new_node);
                cinfo.exit[0] = neighbor_exit[0];
                cinfo.exit[1] = neighbor_exit[1];
                neighbor->children.push_back(cinfo);
                (*area)[loc[0]] |= UINT64_C(1) << loc[1];
                node_map[loc[0]][loc[1]] = new_node;
            }
        }
        explore->expand();
        explore->andnot(*area);
        *explore &= terr;
    }
    explore = Territory::free(explore);

    collapse_doubled(root);
    root->set_size();
}

Tree::~Tree()
{
    root = Node::free(root);
    area = Territory::free(area);
    node_map = Tree::free_map(node_map);
}

std::string& Tree::toString()
{
    static const char* idents = "0123456789ABCDEFGHIJKLMNOPQRSTUVWQYZabcdefghijklmnopqrstuvwxyz";
    Node* idmap[1000];
    int num_mapped = 0;
    std::string* out = new std::string();
    for (int y=0; y < Map::map_height; y++)
    {
        for (int x=0; x < Map::map_width; x++)
        {
            Node* current = node_map[x][y];
            if (current != NULL)
            {
                int id = -1;
                while (id < num_mapped)
                {
                    if (idmap[++id] == current)
                        break;
                }
                if (id == num_mapped)
                {
                    idmap[id] = current;
                    ++num_mapped;
                    num_mapped = num_mapped < 999 ? num_mapped : 999;
                }
                *out += idents[id % 62];
                for (int i=0; i < 1000; i++)
                {
                    if (current == root || current == NULL)
                        break;
                    current = current->parent;
                }
                if (current != root)
                {
                    fprintf(stderr, "Found node that does not go to the root\n");
                }
            } else {
                out->append(".");
            }
        }
        out->append("\n");
    }
    std::ostringstream oss;
    oss << "id: size total\n";
    for (int i=0; i < num_mapped; i++)
    {
        int size = idmap[i]->area->size(idmap[i]->entrance);
        oss << std::setiosflags(std::ios::right);
        oss << std::setw(3) << idents[i % 62] << ": ";
        oss << std::setw(3) << size << " ";
        oss << std::setw(5) << idmap[i]->my_size << "\n";
        //oss << idmap[i]->area->toString();
    }
    out->append(oss.str());
    return *out;
}

#define TS_CACHE_SIZE 100000

TreeSizeCache::TreeSizeCache()
{
    hashes = new uint64_t[TS_CACHE_SIZE];
    sizes = new int[TS_CACHE_SIZE];
    bounds = new Bound[TS_CACHE_SIZE];
    for (int i=0; i < TS_CACHE_SIZE; i++)
    {
        hashes[i] = 0;
        sizes[i] = -1;
        bounds[i] = UPPER;
    }
}

int TreeSizeCache::get_size(Territory& area, int* start, Bound& bound)
{
    uint64_t area_hash = UINT64_C(0);
    for (int i=0; i < Map::map_width; i++)
    {
        for (int j=0; j < Map::map_height; j++)
        {
            if (!(area[i] & (UINT64_C(1) << j)))
                area_hash ^= Map::hashkeys_walls[(i * Map::map_height) + j];
        }
    }
    area_hash ^= Map::hashkeys_bikes[(start[0]*Map::map_height) + start[1]];
    int key = area_hash % TS_CACHE_SIZE;
    if (hashes[key] != area_hash || sizes[key] == 1)
    {
        Tree tree = Tree(area, start);
        sizes[key] = tree.root->my_size;
        bounds[key] = tree.root->my_bound;
        hashes[key] = area_hash;
    }
#ifdef CHECK_CACHE
    else {
        Tree checktree = Tree(area, start);
        if (sizes[key] != checktree.root->my_size
                || bounds[key] != checktree.root->my_bound)
        {
            fprintf(stderr, "hash collision in tree cache");
            exit(1);
        }
    }
#endif

    bound = bounds[key];
    return sizes[key];
}

} // namespace

