
#include "territory.h"
#include "Map.h"
#include <cstring>

int popcount(uint64_t value)
{
    const uint64_t b1 = UINT64_C(0x5555555555555555);
    const uint64_t b2 = UINT64_C(0x3333333333333333);
    const uint64_t b4 = UINT64_C(0x0f0f0f0f0f0f0f0f);
    const uint64_t b7 = UINT64_C(0x0101010101010101);

    value -= (value >> 1) & b1;
    value = (value & b2) + ((value >> 2) & b2);
    value = (value + (value >> 4)) & b4;
    value = (value * b7) >> 56;
    return (int)value;
}

#ifdef _MSC_VER

#include <intrin.h>

#ifdef _WIN64

int bitindex(uint64_t bit)
{
    unsigned long index;
    if (_BitScanForward64(&index, bit))
        return index;
    else
        return 0;
}

#else

int bitindex(uint64_t bit)
{
    unsigned long index;
    if (_BitScanForward(&index, (unsigned long)bit))
    {
        return index;
    } else {
        if (_BitScanForward(&index, (unsigned long)(bit >> 32)))
            return index + 32;
        else
            return 0;
    }
}

#endif

#else

int bitScanForward(uint64_t x) {
   asm ("bsfq %0, %0" : "=r" (x) : "0" (x));
   return (int) x;
}

int bitindex(uint64_t bit)
{
    if (bit)
        return bitScanForward(bit);
    else
        return 0;
}

/*
int bitindex(uint64_t bit)
{
    int index = ((bit & UINT64_C(0xAAAAAAAAAAAAAAAA)) != 0);
    index |= ((bit & UINT64_C(0xCCCCCCCCCCCCCCCC)) != 0) << 1;
    index |= ((bit & UINT64_C(0xF0F0F0F0F0F0F0F0)) != 0) << 2;
    index |= ((bit & UINT64_C(0xFF00FF00FF00FF00)) != 0) << 3;
    index |= ((bit & UINT64_C(0xFFFF0000FFFF0000)) != 0) << 4;
    index |= ((bit & UINT64_C(0xFFFFFFFF00000000)) != 0) << 5;
    return index;
}
*/

#endif

#define TERRITORY_CACHE_SIZE 1000
std::vector<Territory*> Territory::cache;
#ifdef MEMORY_CHECK
int Territory::allocated = 0;

int Territory::unreturned()
{
    return allocated - cache.size();
}
#endif


void Territory::preallocate()
{
    for (int i=0; i < TERRITORY_CACHE_SIZE; i++)
    {
        cache.push_back(new Territory());
#ifdef MEMORY_CHECK
        ++allocated;
#endif
    }
}

Territory* Territory::allocate()
{
    Territory* t;
    if (cache.size() != 0)
    {
        t = cache.back();
        cache.pop_back();
    } else {
        t = new Territory();
#ifdef MEMORY_CHECK
        ++allocated;
#endif
    }
    t->clear();
    return t;
}

Territory* Territory::allocate(const Territory& other)
{
    Territory* t;
    if (cache.size() != 0)
    {
        t = cache.back();
        cache.pop_back();
    } else {
        t = new Territory();
#ifdef MEMORY_CHECK
        ++allocated;
#endif
    }
    (*t) = other;
    return t;
}

Territory* Territory::free(Territory* territory)
{
    cache.push_back(territory);
    return NULL;
}

Territory::Territory()
{
    area = new uint64_t[Map::map_width];
}

Territory::~Territory()
{
    delete[] area;
}

void Territory::clear()
{
    memset(area, 0, sizeof(uint64_t) * Map::map_width);
}

std::string& Territory::toString()
{
    std::string* board = new std::string();
    for (int y=0; y < Map::map_height; y++)
    {
        for (int x=0; x < Map::map_width; x++)
        {
            if (area[x] & (UINT64_C(1) << y))
            {
                board->append("#");
            } else {
                board->append(".");
            }
        }
        board->append("\n");
    }
    return *board;
}

Territory& Territory::operator=(const Territory& other)
{
    memcpy(area, other.area, sizeof(uint64_t) * Map::map_width);
    return *this;
}

Territory& Territory::operator|=(const Territory& other)
{
    for (int i=0; i < Map::map_width; i++)
    {
        area[i] |= other.area[i];
    }
    return *this;
}

Territory& Territory::operator&=(const Territory& other)
{
    for (int i=0; i < Map::map_width; i++)
    {
        area[i] &= other.area[i];
    }
    return *this;
}

uint64_t& Territory::operator[](const int column)
{
    return area[column];
}

bool Territory::operator==(const Territory& other)
{
    if (this == &other)
        return true;
    return memcmp(area, other.area, sizeof(uint64_t) * Map::map_width) == 0;
}

bool Territory::operator!=(const Territory& other)
{
    if (this == &other)
        return false;
    return memcmp(area, other.area, sizeof(uint64_t) * Map::map_width) != 0;
}

bool Territory::not_empty()
{
    for (int i=0; i < Map::map_width; i++)
    {
        if (area[i] != 0)
            return true;
    }
    return false;
}

bool Territory::contains(int* point)
{
    if (point[0] < 0 || point[1] < 0 ||
            point[0] > Map::map_width || point[1] > Map::map_height)
        return false;
    return (area[point[0]] & (UINT64_C(1) << point[1])) != 0;
}

bool Territory::contains(int x, int y)
{
    if (x < 0 || y < 0 || x > Map::map_width || y > Map::map_height)
        return false;
    return (area[x] & (UINT64_C(1) << y)) != 0;
}

#define ODDS UINT64_C(0x5555555555555555)
#define EVENS UINT64_C(0xAAAAAAAAAAAAAAAA)
int Territory::size(int* start)
{
    int start_color;
    int color_size[2];
    color_size[0] = 0;
    color_size[1] = 0;
    for (int i=0; i < Map::map_width; i++)
    {
        uint64_t ws, bs;
        if (i % 2)
        {
            ws = ODDS;
            bs = EVENS;
        } else {
            ws = EVENS;
            bs = ODDS;
        }
        color_size[0] += popcount(area[i] & ws);
        color_size[1] += popcount(area[i] & bs);
    }
    uint64_t sq_color = start[0] % 2 ? ODDS : EVENS;
    start_color = (UINT64_C(1) << start[1]) & sq_color ? 0 : 1;
    int min_color = color_size[0] == color_size[1] ? start_color
        : (color_size[0] < color_size[1] ? 0 : 1);
    int size = color_size[min_color] * 2;
    if (min_color != start_color)
        ++size;
    return size;
}

int Territory::size(int* start, int* end)
{
    int start_color, end_color;
    int color_size[2];
    color_size[0] = 0;
    color_size[1] = 0;
    for (int i=0; i < Map::map_width; i++)
    {
        uint64_t ws, bs;
        if (i % 2)
        {
            ws = ODDS;
            bs = EVENS;
        } else {
            ws = EVENS;
            bs = ODDS;
        }
        color_size[0] += popcount(area[i] & ws);
        color_size[1] += popcount(area[i] & bs);
    }
    uint64_t sq_color = start[0] % 2 ? ODDS : EVENS;
    start_color = (UINT64_C(1) << start[1]) & sq_color ? 0 : 1;
    sq_color = end[0] % 2 ? ODDS : EVENS;
    end_color = (UINT64_C(1) << end[1]) & sq_color ? 0 : 1;

    int size = 0;
    int min_color;
    if (color_size[0] == color_size[1])
    {
        min_color = 0;
        if (start_color == end_color)
            --size;
    } else {
        min_color = color_size[0] < color_size[1] ? 0 : 1;
        if (min_color != start_color)
            ++size;
        if (min_color == end_color)
            --size;
    }
    size += color_size[min_color] * 2;
    return size;
}


int Territory::neighbors(int* point)
{
    int count = 0;
    uint64_t ybit = UINT64_C(1) << point[1];
    if (point[0] > 0)
    {
        bool west = (area[point[0]-1] & ybit) == 0;
        count = west ? 0 : 1;
    }
    if (point[0] < Map::map_width - 1)
    {
        bool east = (area[point[0]+1] & ybit) == 1;
        count += east ? 0 : 1;
    }
    count += popcount(area[point[0]] & ((ybit << 1) | (ybit >> 1)));
    return count;
}

int Territory::space_available(int* start)
{
    Territory* space = allocate();
    Territory* previous = allocate();
    space->area[start[0]] = UINT64_C(1) << start[1];
    while (*space != *previous)
    {
        *previous = *space;
        space->expand();
        *space &= *this;
    }
    int size = space->size(start);
    free(previous);
    free(space);
    return size;
}

bool Territory::separate(int* p1start, int* p2start,
        Territory& p1terr, Territory& p2terr)
{
    Territory* p1prev = allocate();
    Territory* p2prev = allocate();
    Territory* shared = allocate();

    p1terr.area[p1start[0]] |= UINT64_C(1) << p1start[1];
    p2terr.area[p2start[0]] |= UINT64_C(1) << p2start[1];
    bool have_shared = false;
    bool p1update = true;
    bool p2update = true;
    while (p1update || p2update)
    {
        if (p1update)
            *p1prev = p1terr;
        if (p2update)
            *p2prev = p2terr;
        if (p1update)
        {
            p1terr.expand();
            p1terr &= *this;
            if (p1terr.overlap(*p2prev))
            {
                p1terr.andnot(*p2prev);
                have_shared = true;
            }
        }
        if (p2update)
        {
            p2terr.expand();
            p2terr &= *this;
            if (p2terr.overlap(*p1prev))
            {
                p2terr.andnot(*p1prev);
                have_shared = true;
            }
        }
        if (shared->set_and_clear_overlap(p1terr, p2terr))
            have_shared = true;
        p1update = p1terr != *p1prev;
        p2update = p2terr != *p2prev;
    }
    free(p1prev);
    free(p2prev);
    free(shared);
    return have_shared;
}

bool Territory::overlap(const Territory& other)
{
    for (int i=0; i < Map::map_width; i++)
    {
        if (area[i] & other.area[i])
            return true;
    }
    return false;
}

void Territory::andnot(const Territory& other)
{
    for (int i=0; i < Map::map_width; i++)
    {
        area[i] &= ~other.area[i] & Map::map_mask;
    }
}

void Territory::expand()
{
    uint64_t next_row = 0;
    for (int i=0; i < Map::map_width; i++)
    {
        if(i > 0)
            area[i-1] |= area[i];
        uint64_t nterr = area[i] | area[i] << 1 | area[i] >> 1
            | next_row;
        next_row = area[i];
        area[i] |= nterr;
    }
}

bool Territory::set_and_clear_overlap(Territory& p1terr, Territory& p2terr)
{
    bool have_shared = false;
    for (int i=0; i < Map::map_width; i++)
    {
        area[i] |= p1terr.area[i] & p2terr.area[i];
        if (area[i])
            have_shared = true;
        p1terr.area[i] &= ~area[i] & Map::map_mask;
        p2terr.area[i] &= ~area[i] & Map::map_mask;
    }
    return have_shared;
}

void Territory::fill_from(const Territory& other, int* start)
{
    Territory* previous = allocate();
    area[start[0]] = UINT64_C(1) << start[1];
    while (*this != *previous)
    {
        *previous = *this;
        expand();
        *this &= other;
    }
    free(previous);
}

TerritoryIter::TerritoryIter(Territory* t) : x(0)
{
    terr = t;
    column = (*terr)[0];
}

bool TerritoryIter::has_more()
{
    if (column == 0)
        next_col();
    return column != 0;
}

int* TerritoryIter::next()
{
    if (column == 0)
    {
        if (!next_col())
        {
            current[0] = -1;
            current[1] = -1;
        }
    }

    current[0] = x;
    uint64_t bit = column & -column;
    column ^= bit;
    current[1] = bitindex(bit);
    return current;
}

bool TerritoryIter::next_col()
{
    bool found_more = false;
    while (x < Map::map_width-1)
    {
        column = (*terr)[++x];
        if (column != 0)
        {
            found_more = true;
            break;
        }
    }
    return found_more;
}

